@@ -1,15 +1,21 @@
 import torch
 import torch.nn as nn
 import torch.nn.functional as F
-from functools import partial
+import math
+from operator import mul
+from functools import partial, reduce
 from routing_transformer.reversible import ReversibleSequence, SequentialSequence
 
 # constants
 
 TOKEN_SELF_ATTN_VALUE = -5e4
+KMEAN_INIT_ITERS = 10
 
 # helper functions
 
+def identity(x, *args, **kwargs):
+    return x
+
 def default(val, default_val):
     return default_val if val is None else val
 
@@ -76,14 +82,14 @@ def kmeans_iter(x, means, num_clusters, compute_se_):  # hard k-means single ite
     se = compute_se(x, means) if compute_se_ else None
     return means, se
 
-def kmeans(x, means, training=True, reset=False, compute_se=False):
+def kmeans(x, means, training=True, init=False, compute_se=False):
     b, *_, d = x.shape
     num_clusters = means.shape[1]
 
     max_iters = 1 if training else 0
     
-    if reset:
-        max_iters = max(10, max_iters)
+    if init:
+        max_iters = max(KMEAN_INIT_ITERS, max_iters)
         means = x.transpose(0, 1).contiguous().view(t, -1, d)
         means = means[:, torch.randperm(num_clusters, device=x.device)[:num_clusters]]
 
@@ -122,6 +128,19 @@ def split_at_index(dim, index, t):
 
 # helper classes
 
+class Chunk(nn.Module):
+    def __init__(self, chunks, fn, along_dim = -1):
+        super().__init__()
+        self.dim = along_dim
+        self.chunks = chunks
+        self.fn = fn
+
+    def forward(self, x):
+        if self.chunks <= 1:
+            return self.fn(x)
+        chunks = x.chunk(self.chunks, dim = self.dim)
+        return torch.cat([self.fn(c) for c in chunks], dim = self.dim)
+
 class PreNorm(nn.ModuleList):
     def __init__(self, dim, fn):
         super().__init__()
@@ -162,16 +181,66 @@ class RelativePositionalEmbedding(nn.Module):
         emb = torch.einsum('bhnid,jhd->bhnij', q, self.weights.type(q.dtype)) * self.scale
         return shift(emb)
 
+class AxialPositionalEmbedding(nn.Module):
+    def __init__(self, dim, max_seq_len, axial_shape = ()):
+        super().__init__()
+        assert reduce(mul, axial_shape, 1) == max_seq_len, 'axial position shape must multiply up to max sequence length'
+
+        self.dim = dim
+        self.seq_len = max_seq_len
+        self.shape = axial_shape
+
+        self.weights = ParameterList(self, 'weights', len(axial_shape))
+
+        for ind, shape in enumerate(self.shape):
+            ax_shape = [1] * len(self.shape)
+            ax_shape[ind] = shape
+            ax_shape = (1, *ax_shape, dim)
+            ax_emb = nn.Parameter(torch.zeros(ax_shape).normal_(0, 1))
+            self.weights.append(ax_emb)
+
+    def forward(self, x):
+        b, t, e = x.shape
+        embs = []
+
+        for ax_emb in self.weights.to_list():
+            expand_shape = (b, *self.shape, self.dim)
+            emb = ax_emb.expand(expand_shape).reshape(b, self.seq_len, self.dim)
+            embs.append(emb)
+
+        pos_emb = sum(embs)
+        return pos_emb[:, :t].to(x)
+
+# a mock parameter list object until below issue is resolved
+# https://github.com/pytorch/pytorch/issues/36035
+class ParameterList(object):
+    def __init__(self, kls, prefix, length):
+        self.ind = 0
+        self.kls = kls
+        self.prefix = prefix
+        self.length = length
+
+    def _keyname(self, prefix, ind):
+        return f'{prefix}_{ind}'
+
+    def append(self, x):
+        setattr(self.kls, self._keyname(self.prefix, self.ind), x)
+        self.ind += 1
+
+    def to_list(self):
+        return [getattr(self.kls, self._keyname(self.prefix, i)) for i in range(self.length)]
+
 # local attention
 
 class LocalAttention(nn.Module):
-    def __init__(self, bucket_size, heads, head_dim, causal = False, look_backward = 1, look_forward = 0, dropout = 0., shared_qk = False):
+    def __init__(self, bucket_size, heads, head_dim, causal = False, look_backward = 1, look_forward = None, dropout = 0., shared_qk = False):
         super().__init__()
-        assert not (causal and look_forward > 0), 'you cannot look forward if causal'
+        self.look_forward = default(look_forward, 0 if causal else 1)
+        assert not (causal and self.look_forward > 0), 'you cannot look forward if causal'
+
         self.bucket_size = bucket_size
         self.causal = causal
         self.look_backward = look_backward
-        self.look_forward = look_forward
         self.shared_qk = shared_qk
 
         self.heads = heads
@@ -245,8 +314,16 @@ class LocalAttention(nn.Module):
 
 # kmeans attention
 
+def register_kmeans_update_on_backwards(module):
+    module.kmean_attention_modules = find_modules(module, KmeansAttention)
+    def hook(_, grad_in, grad_out):
+        for m in module.kmean_attention_modules:
+            m.update_kmeans()
+
+    return module.register_backward_hook(hook)
+
 class KmeansAttention(nn.Module):
-    def __init__(self, num_clusters, window_size, num_heads, head_dim, causal = False, dropout = 0., ema_decay = 0.8):
+    def __init__(self, num_clusters, window_size, num_heads, head_dim, causal = False, dropout = 0., ema_decay = 0.999):
         super().__init__()
         self.num_heads = num_heads
         self.num_clusters = num_clusters
@@ -255,6 +332,8 @@ class KmeansAttention(nn.Module):
         self.window_size = window_size
         self.causal = causal
 
+        self.router = nn.Parameter(torch.randn(num_heads, head_dim, head_dim))
+
         self.rel_pos = RelativePositionalEmbedding(head_dim, num_heads, window_size)
 
         self.ema_decay = ema_decay
@@ -282,15 +361,14 @@ class KmeansAttention(nn.Module):
         num_clusters = t // wsz
 
         with torch.no_grad():
-            k = F.normalize(qk, dim=-1)
-            reset = not self.initted
+            k_routing = torch.einsum('bhtd,hdr->bhtr', qk, self.router)
+            k_routing = F.normalize(qk, dim=-1)
 
-            means, dists, se = kmeans(k, self.means, training=self.training, reset=reset)
+            means, dists, se = kmeans(k_routing, self.means, training=self.training, init=not self.initted)
             indices = distribution(dists, wsz)
 
             if self.training:
                 self.new_means.copy_(means)
-                # self.update_kmeans(means)
 
             indices = indices.contiguous().view(*indices.size()[:2], -1)
         
@@ -358,7 +436,7 @@ class FeedForward(nn.Module):
 # self attention
 
 class SelfAttention(nn.Module):
-    def __init__(self,  dim, depth, max_seq_len, heads, local_attn_heads, window_size, causal = False, attn_dropout = 0., dropout = 0.):
+    def __init__(self,  dim, depth, max_seq_len, heads, local_attn_heads, window_size, causal = False, attn_dropout = 0., dropout = 0., kmeans_ema_decay = 0.999):
         super().__init__()
         assert (dim % heads) == 0, 'hidden dimension must be divisible by number of heads'
         assert (max_seq_len % window_size) == 0, 'maximum sequence length must be divisible by the target window size'
@@ -378,7 +456,7 @@ class SelfAttention(nn.Module):
             self.local_attn = LocalAttention(window_size // 2, local_attn_heads, head_dim, causal = True, shared_qk = True, dropout = attn_dropout)
 
         if self.global_attn_heads > 0:
-            self.global_attn = KmeansAttention(num_clusters, window_size, self.global_attn_heads, head_dim, causal = causal, dropout = attn_dropout)
+            self.global_attn = KmeansAttention(num_clusters, window_size, self.global_attn_heads, head_dim, causal = causal, dropout = attn_dropout, ema_decay = kmeans_ema_decay)
 
         self.to_qkv = nn.Linear(dim, dim * 2, bias = False)
         self.to_out = nn.Linear(dim, dim, bias = False)
@@ -408,17 +486,8 @@ class SelfAttention(nn.Module):
         out = self.to_out(out)
         return self.dropout(out)
 
-def register_kmeans_update_on_backwards(module):
-    module.kmean_attention_modules = find_modules(module, KmeansAttention)
-
-    def hook(_, grad_in, grad_out):
-        for m in module.kmean_attention_modules:
-            m.update_kmeans()
-
-    module.register_backward_hook(hook)
-
 class RoutingTransformer(nn.Module):
-    def __init__(self, dim, depth, max_seq_len, heads = 8, window_size = 64, causal = False, attn_dropout = 0., ff_dropout = 0., attn_layer_dropout = 0., n_local_attn_heads = 0, ff_glu = False, reversible = False):
+    def __init__(self, dim, depth, max_seq_len, heads = 8, window_size = 64, causal = False, attn_dropout = 0., ff_dropout = 0., attn_layer_dropout = 0., layer_dropout = 0., n_local_attn_heads = 0, ff_glu = False, reversible = False, ff_chunks = 1, kmeans_ema_decay = 0.999):
         super().__init__()
         if type(n_local_attn_heads) is not tuple:
             n_local_attn_heads = tuple([n_local_attn_heads] * depth)
@@ -427,36 +496,36 @@ class RoutingTransformer(nn.Module):
         assert all([local_heads <= heads for local_heads in n_local_attn_heads]), 'number of local attn heads must be less than the maximum number of heads'
 
         layers = nn.ModuleList([])
+        fn_wrapper = partial(PreNorm, dim)
+
         for ind, local_heads in zip(range(depth), n_local_attn_heads):
-            attn = SelfAttention(dim, depth, max_seq_len, heads, local_heads, window_size, causal = causal, attn_dropout = attn_dropout, dropout = attn_layer_dropout)
-            ff = FeedForward(dim, dropout = ff_dropout, glu = ff_glu)
-
-            attn = PreNorm(dim, attn)
-            ff = PreNorm(dim, ff)
+            attn = SelfAttention(dim, depth, max_seq_len, heads, local_heads, window_size, causal = causal, attn_dropout = attn_dropout, dropout = attn_layer_dropout, kmeans_ema_decay = kmeans_ema_decay)
+            ff = Chunk(ff_chunks, FeedForward(dim, dropout = ff_dropout, glu = ff_glu), along_dim=1)
 
+            attn, ff = map(fn_wrapper, (attn, ff))
             layers.append(nn.ModuleList([attn, ff]))
 
         execute_type = ReversibleSequence if reversible else SequentialSequence
-        self.layers = execute_type(layers)
+        self.layers = execute_type(layers, layer_dropout = layer_dropout)
         self.pad_to_window_size = window_size
 
-        register_kmeans_update_on_backwards(self)        
+        self.remove_handle_fn = register_kmeans_update_on_backwards(self)        
 
     def forward(self, x, **kwargs):
         x = self.layers(x, **kwargs)
         return x
 
 class RoutingTransformerLM(nn.Module):
-    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, window_size = 64, causal = False, attn_dropout = 0., ff_dropout = 0., attn_layer_dropout = 0., ff_mult = 4, ff_activation = None, ff_glu = False, post_attn_dropout = 0., return_embeddings = False, n_local_attn_heads = 0, reversible = False):
+    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, window_size = 64, causal = False, attn_dropout = 0., ff_dropout = 0., attn_layer_dropout = 0., layer_dropout = 0., ff_mult = 4, ff_activation = None, ff_glu = False, return_embeddings = False, n_local_attn_heads = 0, reversible = False, ff_chunks = 1, kmeans_ema_decay = 0.999):
         super().__init__()
         self.max_seq_len = max_seq_len
         self.token_emb = nn.Embedding(num_tokens, dim)
-        self.pos_emb = AbsolutePositionalEmbedding(dim, max_seq_len)
-        self.routing_transformer = RoutingTransformer(dim, depth, max_seq_len, heads = heads, window_size = window_size, causal = causal, ff_dropout = ff_dropout, attn_dropout = attn_dropout, attn_layer_dropout = attn_layer_dropout, n_local_attn_heads = n_local_attn_heads, ff_glu = ff_glu, reversible = reversible)
-        self.out = nn.Linear(dim, num_tokens)
+        self.axial_pos_emb = AxialPositionalEmbedding(dim, max_seq_len, axial_shape=(max_seq_len // window_size, window_size))
+        self.routing_transformer = RoutingTransformer(dim, depth, max_seq_len, heads = heads, window_size = window_size, causal = causal, ff_dropout = ff_dropout, attn_dropout = attn_dropout, attn_layer_dropout = attn_layer_dropout, layer_dropout = layer_dropout, n_local_attn_heads = n_local_attn_heads, ff_glu = ff_glu, reversible = reversible, ff_chunks = ff_chunks, kmeans_ema_decay = kmeans_ema_decay)
+        self.out = nn.Linear(dim, num_tokens) if not return_embeddings else identity
 
     def forward(self, x, **kwargs):
         x = self.token_emb(x)
-        x = x + self.pos_emb(x).type(x.type())
+        x = x + self.axial_pos_emb(x)
         x = self.routing_transformer(x, **kwargs)
         return self.out(x)
